// Freeslot and asset definition script

freeslot(
	// Sprites
	"SPR_THUN",
	"SPR_THUF",
	
	// Player states
	"S_PLAY_PIKAFLIP",
	"S_PLAY_CHUJUMPFALL",
	"S_PLAY_VOLTTACKLE",
	"S_PLAY_IRONTAIL_ANTIC",
	"S_PLAY_IRONTAIL_LAUNCH",
	"S_PLAY_IRONTAIL",
	"S_PLAY_THUNDER",
	"S_PLAY_DOUBLETEAM",
	"S_PLAY_CHUPARRY",
	"S_PLAY_CHUPARRY_FINISH",
	"S_PLAY_CHUVICTORY",
	"S_PLAY_CHUVICTORY_FINISH",
	"S_PLAY_DEADCHU",
	"S_PLAY_DEADCHU_FINISH",
	
	// States for decorative objects
	"S_VOLTTACKLE_SPARK",
	"S_CHUTHUNDER",
	"S_CHUTHUNDER_BOLT",
	"S_DOUBLETEAM",
	
	// Decorative objects
	"MT_DOUBLETEAM",
	"MT_CHUTHUNDER",
	"MT_CHUTHUNDER_BOLT",
	"MT_VOLTTACKLE_SPARK",
	
	// Sound effects
	"sfx_pkjump",
	"sfx_pkhurt",
	"sfx_pkhnve",
	"sfx_pkfain",
	"sfx_pkiron",
	"sfx_pkvolt",
	"sfx_pkfizl",
	"sfx_pkdoub",
	"sfx_pkthun",
	"sfx_pkfein",
	"sfx_pkslam",
	"sfx_pkspkl",
	
	// Voice clips
	"sfx_pvname",
	"sfx_pvdie1",
	"sfx_pvdie2",
	"sfx_pvyay1",
	"sfx_pvyay2",
	"sfx_pvfall",
	"sfx_pvow1",
	"sfx_pvow2",
	"sfx_pvatk1",
	"sfx_pvatk2",
	"sfx_pvatk3",
	"sfx_pvatk4"
)

// Action that launches Pikachu forward during Iron Tail
local A_IronTailLaunch = function(mo,a,b)
	P_InstaThrust(mo,mo.angle,60*FU)
end

// --States

// Player state used in multiple situations, such as running out of Wild Charge
states[S_PLAY_PIKAFLIP] = {
	sprite = SPR_PLAY,
	frame = SPR2_ROLL|FF_SPR2ENDSTATE,
	tics = 2,
	var1 = S_PLAY_FALL,
	nextstate = S_PLAY_PIKAFLIP
}

// Player state used when falling after jumpiung
states[S_PLAY_CHUJUMPFALL] = {
	sprite = SPR_PLAY,
	frame = SPR2_GLID,
	tics = -1
}

// Player state used when using Wild Charge
states[S_PLAY_VOLTTACKLE] = {
	sprite = SPR_PLAY,
	frame = SPR2_FLY|FF_FULLBRIGHT,
	tics = 2,
	nextstate = S_PLAY_VOLTTACKLE
}

// Player state used when winding up for Iron Tail
states[S_PLAY_IRONTAIL_ANTIC] = {
	sprite = SPR_PLAY,
	frame = SPR2_MLEL|FF_FULLBRIGHT,
	tics = 5,
	nextstate = S_PLAY_IRONTAIL_LAUNCH
}

// Action state used to launch player
states[S_PLAY_IRONTAIL_LAUNCH] = {
	sprite = SPR_PLAY,
	frame = SPR2_MLEE,
	action = A_IronTailLaunch,
	tics = 0,
	nextstate = S_PLAY_IRONTAIL
}

// Player state used when using Iron Tail
states[S_PLAY_IRONTAIL] = {
	sprite = SPR_PLAY,
	frame = SPR2_MLEE,
	tics = 1,
	nextstate = S_PLAY_IRONTAIL
}

// Player state used when using Thunder
states[S_PLAY_THUNDER] = {
	sprite = SPR_PLAY,
	frame = SPR2_BNCE|FF_FULLBRIGHT,
	tics = 1,
	nextstate = S_PLAY_THUNDER
}

// Player state used when using Double Team
states[S_PLAY_DOUBLETEAM] = {
	sprite = SPR_NULL,
	tics = 15,
	nextstate = S_PLAY_FALL
}

// Player states used when beating a level
states[S_PLAY_CHUVICTORY] = {
	sprite = SPR_PLAY,
	frame = SPR2_SWIM|FF_SPR2ENDSTATE,
	tics = 4,
	var1 = S_PLAY_CHUVICTORY_FINISH,
	nextstate = S_PLAY_CHUVICTORY
}
states[S_PLAY_CHUVICTORY_FINISH] = {
	sprite = SPR_PLAY,
	frame = SPR2_CLNG,
	action = A_PlaySound,
	var1 = sfx_pkspkl,
	var2 = 1,
	tics = -1
}

// Player states used when parrying an attack with Double Team
states[S_PLAY_CHUPARRY] = {
	sprite = SPR_PLAY,
	frame = SPR2_SWIM|FF_SPR2ENDSTATE,
	tics = 3,
	var1 = S_PLAY_CHUPARRY_FINISH,
	nextstate = S_PLAY_CHUPARRY
}
states[S_PLAY_CHUPARRY_FINISH] = {
	sprite = SPR_PLAY,
	frame = SPR2_CLMB,
	tics = TICRATE/2,
	nextstate = S_PLAY_PIKAFLIP
}

// Player states used for custom death animation
states[S_PLAY_DEADCHU] = {
	sprite = SPR_PLAY,
	frame = SPR2_FIRE|FF_SPR2ENDSTATE,
	tics = 2,
	var1 = S_PLAY_DEADCHU_FINISH,
	nextstate = S_PLAY_DEADCHU
}
states[S_PLAY_DEADCHU_FINISH] = {
	sprite = SPR_PLAY,
	frame = SPR2_TWIN,
	tics = -1
}



// State used for Double Team copies
states[S_DOUBLETEAM] = {
	sprite = SPR_PLAY,
	frame = SPR2_LAND,
	tics = -1
}

// State used for Wild Charge sparks
states[S_VOLTTACKLE_SPARK] = {
	sprite = SPR_MCSP,
	tics = 5
}

// State used for Thunder field
states[S_CHUTHUNDER] = {
	sprite = SPR_THUF,
	frame = A|FF_ANIMATE|FF_PAPERSPRITE,
	var1 = 3,
	var2 = 2,
	tics = 8
}

// State used for Thunder bolt (not to be confused with Thunderbolt)
states[S_CHUTHUNDER_BOLT] = {
	sprite = SPR_THUN,
	frame = A|FF_ANIMATE,
	var1 = 3,
	var2 = 1,
	tics = 4
}



// --Mobj types

// Wild Charge spark
mobjinfo[MT_VOLTTACKLE_SPARK] = {
	spawnstate = S_VOLTTACKLE_SPARK,
	radius = 1*FU,
	height = 1*FU,
	flags = MF_SCENERY|MF_NOBLOCKMAP|MF_NOCLIP|MF_NOCLIPHEIGHT|MF_NOGRAVITY
}

// Thunder wave
mobjinfo[MT_CHUTHUNDER] = {
	spawnstate = S_CHUTHUNDER,
	radius = 1*FU,
	height = 1*FU,
	flags = MF_SCENERY|MF_NOBLOCKMAP|MF_NOCLIP|MF_NOCLIPHEIGHT|MF_NOGRAVITY
}

// Thunder bolt
mobjinfo[MT_CHUTHUNDER_BOLT] = {
	spawnstate = S_CHUTHUNDER_BOLT,
	radius = 1*FU,
	height = 256*FU,
	flags = MF_SCENERY|MF_NOBLOCKMAP|MF_NOCLIP|MF_NOCLIPHEIGHT|MF_NOGRAVITY
}

// Double Team doubles
mobjinfo[MT_DOUBLETEAM] = {
	spawnstate = S_DOUBLETEAM,
	radius = 16*FU,
	height = 48*FU,
	flags = MF_SCENERY|MF_NOBLOCKMAP|MF_NOCLIP|MF_NOCLIPHEIGHT|MF_NOGRAVITY
}

// --Captions
sfxinfo[sfx_pkhurt].caption = "Pain"
sfxinfo[sfx_pkfain].caption = "Fainting"
sfxinfo[sfx_pkiron].caption = "Iron Tail"
sfxinfo[sfx_pkdoub].caption = "Double Team"
sfxinfo[sfx_pkvolt].caption = "Volt Tackle"
sfxinfo[sfx_pkjump].caption = "Jump"
sfxinfo[sfx_pkfizl].caption = "Volt Tackle fizzles"
sfxinfo[sfx_pkthun].caption = "Thunder"
sfxinfo[sfx_pkfein].caption = "Parry"
sfxinfo[sfx_pkslam].caption = "Thud"
sfxinfo[sfx_pkhnve].caption = "Ineffective hit"
sfxinfo[sfx_pkspkl].caption = "Audible joy"

sfxinfo[sfx_pvname].caption = "Pikachu!"
sfxinfo[sfx_pvdie1].caption = " "
sfxinfo[sfx_pvdie2].caption = " "
sfxinfo[sfx_pvyay1].caption = " "
sfxinfo[sfx_pvyay2].caption = " "
sfxinfo[sfx_pvfall].caption = " "
sfxinfo[sfx_pvow1].caption = " "
sfxinfo[sfx_pvow2].caption = " "
sfxinfo[sfx_pvatk1].caption = " "
sfxinfo[sfx_pvatk2].caption = " "
sfxinfo[sfx_pvatk3].caption = " "
sfxinfo[sfx_pvatk4].caption = " "