// Main logic script

local usevox = CV_FindVar("pikavoices")
local lowParticles = CV_FindVar("pikareduceparticles")

// Some variables that affect various parts of various moves
local wcMaxCharge = TICRATE
local thunderRadius = 150
local dtParryWindow = 5
local dtCooldown = TICRATE



// Checks whether or not the player should use Pikachu logic
local isChu = function(p)
	if p.mo and p.mo.valid
	and p.mo.skin == "pikachu"
		return true end
	return false
end

// Checks whether Pikachu can use moves
local chuCanAct = function(p)
	if isChu(p) // Must be Pikachu
	and p.chuTable // Must have a defined Pikachu variable table
	and (P_IsObjectOnGround(p.mo) or p.pflags & PF_JUMPED) // Must be jumping or on the ground
	// Must not...
	and not (p.exiting // ...be finishing the level,
	or p.powers[pw_nocontrol] // have their controls locked,
	or p.powers[pw_carry] // be getting carried by something,
	or P_PlayerInPain(p) // be receiving knockback,
	or p.chuTable.usingSomething // already be using a Pikachu ability,
	or (p.pflags & PF_SPINNING) // be spinning,
	or p.chuTable.sliding // be on a waterslide,
	or p.chuTable.thunderState // be using Thunder or recovering from it,
	or p.playerstate != PST_LIVE) // or be dead.
		return true end
	return false
end

// Sets up the Pikachu variable table
local initChu = function() return {
	usingSomething = false, // Whether or not the player is using a Pikachu move
	voltTackleDuration = 0, // How much longer the player can use Volt Tackle for
	vtSupression = false, // Whether the player must release Jump to begin using Volt Tackle
	thunderState = 0, // Whether the player is using thunder, or if they're bouncing on the ground from it
	thunderTimer = 0, // How long the player has been using Thunder (used to avoid softlocks)
	doubleTeamCooldown = 0, // How long before Pikachu can use Double Team again
	sliding = 0, // Whether or not the player is using a waterslide (since PF_SLIDING isn't available after PreThinkFrame)
	lastButtons = 0, // What buttons the player pressed last frame
	zapped = false, // Whether or not the player's Electric-type resistance was used
	victoryTaunted = false // Whether or not the player has done their victory animation
} end

// Spawns a double for Double Team
local makeDouble = function(mo)
	local double = P_SpawnMobj(mo.x,mo.y,mo.z,MT_DOUBLETEAM)
	double.target = mo // Make it follow the player
	double.skin = mo.skin
	double.angle = mo.player.drawangle
	double.eflags = mo.eflags // Make it flipped if the player is flipped
	double.height = mo.height // Make sure it doesn't clip into the ground if flipped
	return double
end

// Returns whether or not a player mobj could parry an incoming attack with Double Team this tic
local canDoubleTeamParry = function(mo)
	if mo.state == S_PLAY_DOUBLETEAM
	and mo.tics > states[S_PLAY_DOUBLETEAM].tics - dtParryWindow
		return true end
		
	return false
end

// Draw in rings, because for some reason you can't just do that from the player's end with vanilla functions
local attractRings = function(p)

	// Don't attract rings during i-frames, to avoid being able to just instantly pick up lost rings
	if p.powers[pw_flashing]
		return end
		
	local mo = p.mo
	local rad = 256*mo.scale
	
	searchBlockmap("objects",function(r,f)
		// Only work on rings
		if f.type != MT_RING
		and f.type != MT_FLINGRING
		and f.type != MT_REDTEAMRING
		and f.type != MT_BLUETEAMRING
			return nil end
			
		// Spherical hitbox. Hitball? Hit-orb?
		if FixedHypot((f.z+f.height/2)-(r.z+r.height/2),R_PointToDist2(r.x,r.y,f.x,f.y)) <= rad + f.radius
			P_HomingAttack(f,r)
		end
	end,mo,mo.x-rad,mo.x+rad,mo.y-rad,mo.y+rad)
end

// Main 'Chu logic hook
addHook("PlayerThink",function(p)
	if not isChu(p)
		if p.chuTable // Remove Pikachu variables from non-Pikachu players
			p.chuTable = nil
		end
		return // Don't execute any further Pikachu logic
	end
	
	// Initialize Pikachu variables, if they aren't already
	if not p.chuTable
		p.chuTable = initChu()
	end
	
	// Some quick access variables
	local mo = p.mo
	local cmd = p.cmd
	local chu = p.chuTable
	
	// Cancel Thunder while in goop (to prevent cheesing THZ by spamming Thunder) and while watersliding
	if chu.sliding
	or (mo.eflags & MFE_GOOWATER
	and mo.eflags & MFE_UNDERWATER)
		chu.thunderState = 2
		if mo.state == S_PLAY_THUNDER
			mo.state = S_PLAY_FALL
		end
	end
	
	// Upon landing
	if P_IsObjectOnGround(mo)
	or p.powers[pw_carry]
	
		// If the player needs to bounce off the ground with Thunder
		if chu.thunderState == 1
		and not p.powers[pw_carry]
		and p.playerstate == PST_LIVE
		
			chu.thunderState = 2 // Mark player as having bounced
			mo.state = S_PLAY_PIKAFLIP
			attractRings(p) // Attract rings one more time, so that using Thudner on the ground still works
			P_SetObjectMomZ(mo,5*FU,false)
			S_StartSound(mo,sfx_pkslam)
			
			if not lowParticles.value
				for i = 0, 360, 360/16 // Dust particles
					local dust = P_SpawnMobj(mo.x + 64*cos(i*ANG1), mo.y + 64*sin(i*ANG1), mo.z, MT_DUST)
				end
			end
		else
			chu.thunderState = 0
			chu.voltTackleDuration = wcMaxCharge
		end
	
	// Cancel Thunder if the player has exited the proper state for it
	elseif chu.thunderState == 1
	and mo.state != S_PLAY_THUNDER
		chu.thunderState = 0
	end
	
	// Remove pw_strong and mark the player as not using something if they aren't, well, using something
	if mo.state != S_PLAY_DOUBLETEAM
	and mo.state != S_PLAY_IRONTAIL_ANTIC
	and mo.state != S_PLAY_IRONTAIL_LAUNCH
	and mo.state != S_PLAY_IRONTAIL
	and mo.state != S_PLAY_THUNDER
		chu.usingSomething = false
		p.powers[pw_strong] = 0
	end
	
	if mo.state == S_PLAY_THUNDER
		P_SpawnGhostMobj(mo)
		attractRings(p)
	end
	
	if chu.zapped and not p.powers[pw_flashing]
		chu.zapped = false
	end
	
	// If the player is on the ground, sliding, or carried, they must release Jump before starting Volt Tackle
	if P_IsObjectOnGround(mo) or chu.sliding or p.powers[pw_carry]
		chu.vtSupression = true
	elseif ~cmd.buttons & BT_JUMP
		chu.vtSupression = false
	end
	
	// Stop the Double Team sound if not in the proper state
	if S_SoundPlaying(mo,sfx_pkdoub)
	and mo.state != S_PLAY_DOUBLETEAM
		S_StopSoundByID(mo,sfx_pkdoub)
	end
	
	if chu.doubleTeamCooldown
		chu.doubleTeamCooldown = $-1
	end
	
	// Exit out of Thunder if the player has been falling for too long
	if chu.thunderState == 1
		chu.thunderTimer = $+1
		if chu.thunderTimer >= TICRATE*3
			chu.thunderState = 2
			mo.state = S_PLAY_PIKAFLIP
		end
	else
		chu.thunderTimer = 0
	end
	
	// Abilities
	if chuCanAct(p)
	
		// Double Team
		if cmd.buttons & BT_CUSTOM1
		and ~chu.lastButtons & BT_CUSTOM1
		and not chu.doubleTeamCooldown
		
			chu.doubleTeamCooldown = dtCooldown // Put it on cooldown to prevent spam
			chu.usingSomething = true
			
			mo.state = S_PLAY_DOUBLETEAM
			p.powers[pw_strong] = 0 // In case the player uses this to interrupt another move
			S_StopSoundByID(mo,sfx_pkvolt)
			S_StartSound(mo,sfx_pkdoub)
			if usevox.value
				S_StartSound(mo,sfx_pvatk2) // b!
			end
			
			// Doubles
			local double1 = makeDouble(mo)
			double1.movefactor = -32*FU // First double moves to the left
			
			local double2 = makeDouble(mo)
			double2.movefactor = 32*FU // Second double moves to the right...
			double2.flags2 = $|MF2_DONTDRAW // ...and flashes opposite the other one
		
		// Thunder and Iron Tail
		elseif cmd.buttons & BT_SPIN
		and ~chu.lastButtons & BT_SPIN
		and not (mo.eflags & MFE_GOOWATER
		and mo.eflags & MFE_UNDERWATER) // Cannot use while submerged in goop
		
			S_StopSoundByID(mo,sfx_pkvolt)
			
			// Iron Tail
			if cmd.buttons & BT_JUMP
			and mo.state != S_PLAY_IRONTAIL // If this check wasn't there, you could start Iron Tail while already using Iron Tail
			and not P_IsObjectOnGround(mo)
				chu.voltTackleDuration = 0
				
				mo.state = S_PLAY_IRONTAIL_ANTIC // Move winds up for 5 frames before launching
				S_StartSound(mo,sfx_pkiron)
				if usevox.value
					S_StartSound(mo,sfx_pvatk1)
				end
				
				p.powers[pw_strong] = STR_ATTACK|STR_SPIKE|STR_WALL
				
			// Thunder
			else
				P_SetObjectMomZ(mo,-32*FU,false)
				P_InstaThrust(mo,0,0)
				
				mo.state = S_PLAY_THUNDER
				chu.thunderState = 1 // Player is using Thunder
				
				S_StartSound(mo,sfx_pkthun)
				if usevox.value
					if P_IsObjectOnGround(mo)
						S_StartSound(mo,sfx_pvatk2)
					else
						S_StartSound(mo,sfx_pvatk3)
					end
				end
				
				local zaps = 15
				if lowParticles.value
					zaps = 12
				end
				
				// Particles
				for i = 0, 360, 360/zaps
					// Electricity
					local zap = P_SpawnMobj(mo.x + thunderRadius*cos(i*ANG1), mo.y + thunderRadius*sin(i*ANG1), mo.z, MT_CHUTHUNDER)
					zap.angle = R_PointToAngle2(zap.x,zap.y,mo.x,mo.y) + ANGLE_90
					
					// Dust
					P_SpawnMobj(mo.x + thunderRadius*cos(i*ANG1)*3/4, mo.y + thunderRadius*sin(i*ANG1)*3/4, mo.z, MT_SMOKE).scale = 2*FU // Smaller ones further from Pikachu
					if not lowParticles.value
						P_SpawnMobj(mo.x + thunderRadius*cos(i*ANG1)/2, mo.y + thunderRadius*sin(i*ANG1)/2, mo.z, MT_SMOKE).scale = 3*FU // Larger ones closer to Pikachu
					end
				end
				
				// One last visual effect: a bolt of lightning
				P_SpawnMobj(mo.x, mo.y, mo.z, MT_CHUTHUNDER_BOLT)
				
				// A custom P_RadiusAttack that isn't busted (and also destroys spikes and monitors!)
				searchBlockmap("objects",function(r,f)
					if r == f // Trying to shock yourself? Who do you think you are, Pichu?
					or f.player // For some reason, trying to shock a player crashes the game
					or not f.health // You can't kill what's already dead (namely monitors)
						return nil end
						
					if (f.z > r.z
					and R_PointToDist2(r.x,r.y,f.x,f.y) < r.radius) // Smaller, cylindrical hitbox that extends up infinitely
					or FixedHypot((f.z+f.height/2)-(r.z+r.height/2),R_PointToDist2(r.x,r.y,f.x,f.y)) <= (thunderRadius * mo.scale) + f.radius // Larger, spherical hitbox
						
						// Destroy monitors and spikes
						if f.flags & MF_MONITOR
						or f.type == MT_SPIKE
						or f.type == MT_WALLSPIKE
							P_KillMobj(f,r,r,DMG_ELECTRIC)	
							
						// Destroy Egg Guard shields
						elseif f.type == MT_EGGGUARD
						and f.tracer and f.tracer.valid and f.tracer.health
							P_KillMobj(f.tracer,r,r,DMG_ELECTRIC)
							
						// Everything else just gets hurt normally
						elseif f.flags & MF_SHOOTABLE
							P_DamageMobj(f,nil,r,1,DMG_ELECTRIC)	
						end
					end
				end,mo,mo.x-(thunderRadius*mo.scale),mo.x+(thunderRadius*mo.scale),mo.y-(thunderRadius*mo.scale),mo.y+(thunderRadius*mo.scale))
				
				p.powers[pw_strong] = STR_ATTACK|STR_FLOOR|STR_HEAVY|STR_SPRING|STR_SPIKE
				chu.usingSomething = true
			end
			
		// Volt Tackle
		elseif cmd.buttons & BT_JUMP
		and not chu.vtSupression
		and p.pflags & PF_JUMPED
		and chu.voltTackleDuration
			
			local zaps = 64
			if lowParticles.value
				zaps = 16 end
				
			// Spawn zap particles
			for i = 0,zaps
				local angleH = P_RandomKey(360)*ANG1 // Random horizontal angle
				local angleV = P_RandomKey(180)*ANG1 + ANGLE_90 // Random vertical angle
				local rad = 48 // Fixed distance away from Pikachu
				
				// Spawn a zap particle on a sphere around Pikachu using some trigonometry nonsense
				local zap = P_SpawnMobj(mo.x + rad*FixedMul(cos(angleH),cos(angleV)), mo.y + rad*FixedMul(sin(angleH),cos(angleV)),mo.z + mo.height/2 + rad*sin(angleV),MT_VOLTTACKLE_SPARK)
				if lowParticles.value
					zap.scale = 2*FU
				end
				//zap.destscale = 0
				//zap.scalespeed = zap.scale/zap.tics
				//zap.alpha = FU/2
				zap.eflags = mo.eflags // Flip them if the player is flipped
				
				// Make the particles taper off towards the center as they fall
				P_InstaThrust(zap,R_PointToAngle2(zap.x,zap.y,mo.x,mo.y),R_PointToDist2(zap.x,zap.y,mo.x,mo.y)/zap.tics + 4*(P_RandomFixed()-FU/2))
				zap.momz = ((mo.z+mo.height/2)-zap.z)/zap.tics + 4*(P_RandomFixed()-FU/2)
				zap.tics = $+1
			end
			
			// To avoid touchy movement, tend towards existing movement in Automatic mode
			if p.pflags & PF_ANALOGMODE
			and mo.state == S_PLAY_VOLTTACKLE // Not when first starting Volt Tackle
			and abs(mo.momx) + abs(mo.momy) > FU // Not when sitting still
			and not P_BlockingWall(mo, R_PointToAngle2(0,0,mo.momx,mo.momy), 40*FU) // Unless there's a wall there, which would cause you to go off in random directions
				P_InstaThrust(mo,R_PointToAngle2(0,0,mo.momx,mo.momy),40*FU)
			else
				P_InstaThrust(mo,mo.angle,40*FU)
			end
			
			p.powers[pw_strong] = STR_ATTACK|STR_CEILING
			mo.state = S_PLAY_VOLTTACKLE
			attractRings(p)
			
			if not S_SoundPlaying(mo,sfx_pkvolt)
				S_StartSound(mo,sfx_pkvolt)
			end
			
			// Rise slower underwater, otherwise you'd go flying up at ridiculous speeds
			if mo.eflags & MFE_UNDERWATER
				P_SetObjectMomZ(mo,FU/2,true)
			else
				P_SetObjectMomZ(mo,FU,true)
			end
			
			chu.voltTackleDuration = $-1
			if chu.voltTackleDuration == 0
				S_StartSound(mo,sfx_pkfizl)
				mo.state = S_PLAY_PIKAFLIP
			end
		else
			if S_SoundPlaying(mo,sfx_pkvolt)
				S_StopSoundByID(mo,sfx_pkvolt)
			end
			
			if mo.state == S_PLAY_VOLTTACKLE
				mo.state = S_PLAY_ROLL
			end
		end
	elseif S_SoundPlaying(mo,sfx_pkvolt)
		S_StopSoundByID(mo,sfx_pkvolt)
		if mo.state == S_PLAY_VOLTTACKLE
			mo.state = S_PLAY_ROLL
		end
	end
	chu.lastButtons = cmd.buttons // Now that we're done with all of that, keep track of buttons for the next frame
end)

// Input negation and sliding detection
addHook("PreThinkFrame",function()
	for p in players.iterate
		if not p.chuTable // Only run for Pikachu
			continue end
		
		local mo = p.mo
		local cmd = p.cmd
		
		// Can't control yourself while using Double Team or Thunder
		if mo.state == S_PLAY_DOUBLETEAM
		or p.chuTable.thunderState == 1
			cmd.forwardmove = 0
			cmd.sidemove = 0
			
			// ALlow you to keep holding the button that activates whatever move you're using,
			// so that holding it won't count as pressing it once the move finishes
			if mo.state == S_PLAY_DOUBLETEAM
				cmd.buttons = $&BT_CUSTOM1
			elseif mo.state == S_PLAY_THUNDER
				cmd.buttons = $&BT_SPIN
			else
				cmd.buttons = 0
			end
		end
		
		// Freeze in place while winding up for Iron Tail
		if mo.state == S_PLAY_IRONTAIL_ANTIC
			mo.momx = 0
			mo.momy = 0
			mo.momz = 0
		end
		
		// Keep track of PF_SLIDING, since it'll be cleared by the time we get to PlayerThink
		p.chuTable.sliding = p.pflags & PF_SLIDING
	end
end)

// Handle the jump fall state
addHook("PostThinkFrame",function()
	for p in players.iterate()
	
		if p.mo and p.mo.valid
		and p.chuTable
		
		and ((p.mo.state == S_PLAY_FALL
		and p.pflags & PF_JUMPED)
		
		// Prevent airwalking
		or (p.mo.state == S_PLAY_WALK
		or p.mo.state == S_PLAY_RUN
		or p.mo.state == S_PLAY_STND))
		
		and not (P_IsObjectOnGround(p.mo)
		or p.powers[pw_carry])
			p.mo.state = S_PLAY_CHUJUMPFALL
		
		// Don't fall if you're not falling
		elseif p.mo and p.mo.valid
		and P_IsObjectOnGround(p.mo)
		and p.mo.state == S_PLAY_CHUJUMPFALL
			p.mo.state = S_PLAY_STND
		end
	end
end)

// Logic for (not) getting hit, including type resistance and Double Team
addHook("ShouldDamage",function(t,i,s,d,dt)

	if not (t and t.valid
	and t.player and t.player.chuTable)
		return nil end
	
	
	
	if t.player.powers[pw_flashing]
	or t.player.powers[pw_invulnerability]
		return nil
		
	// Don't try to resist or parry an attack from a friend
	elseif s and s.valid and s.player
	and (gametyperules & GTR_FRIENDLY
	or (gametyperules & GTR_TEAMS and s.player.ctfteam == t.player.ctfteam))
		return nil
	end
	
	// Dodging and parrying with Double Team
	if t.state == S_PLAY_DOUBLETEAM
	
		if canDoubleTeamParry(t)
		and s and s.valid
		and s.flags & MF_SHOOTABLE // Don't try and parry things that can't be hurt
			P_Thrust(t,R_PointToAngle2(i.x,i.y,t.x,t.y),16*FU)
			P_SetObjectMomZ(t,8*FU,true)
			
			S_StartSound(t,sfx_pkfein)
			S_StopSoundByID(t,sfx_pkdoub)
			
			if usevox.value
				S_StopSoundByID(t,sfx_pvatk2)
				S_StartSound(mo,sfx_pvatk4)
			end
			
			t.player.powers[pw_flashing] = max($,TICRATE)
			t.state = S_PLAY_CHUPARRY
			
			P_DamageMobj(s,t,t,d,dt)
		end
		
		// No damage
		return false
	end
	
	// Type resistance
	if t.player
	and dt == DMG_ELECTRIC
	and t.player.rings
	and not (t.player.powers[pw_shield] & SH_PROTECTELECTRIC)
		
		// Give extra I-frames when out of rings
		if t.player.rings > 2
			t.player.powers[pw_flashing] = TICRATE/2
			t.player.chuTable.zapped = true
		else
			t.player.powers[pw_flashing] = (TICRATE*3)-1
		end
		
		S_StartSound(t,sfx_pkhnve)
		t.player.losstime = 5*TICRATE // Control how far the rings fly
		P_PlayerRingBurst(t.player,min(t.player.rings,2))
		t.player.rings = $-2
		return false // Don't do regular SRB2 damage
	end
end,MT_PLAYER)

// Double Team double logic
addHook("MobjThinker",function(mo)
	local targ = mo.target
	
	// If the player isn't still there and using Double Team, remove the double
	if not (targ and targ.valid
	and targ.player.chuTable
	and targ.state == S_PLAY_DOUBLETEAM)
		mo.state = S_NULL
	return end
	
	mo.flags2 = $^^MF2_DONTDRAW // Blink
	if canDoubleTeamParry(targ)
		mo.color = SKINCOLOR_WHITE // Flash during the parrying window
		mo.colorized = true
		mo.blendmode = AST_MODULATE
	else
		mo.color = targ.color
		mo.colorized = targ.colorized
		mo.blendmode = targ.blendmode
	end
	
	mo.scale = targ.scale
	mo.destscale = targ.destscale
	mo.eflags = targ.eflags
	
	local off = mo.movefactor
	local ang = ANGLE_90
	if targ.flags2 & MF2_TWOD // Move to the front and back in 2D mode, rather than side to side
		ang = 0
	end
	
	// Move laterally to the player using more trigonometry nonsense
	P_MoveOrigin(mo,targ.x + FixedMul(off,cos(targ.angle+ang)),targ.y + FixedMul(off,sin(targ.angle+ang)),targ.z)
	mo.movefactor = $*3/4
	mo.alpha = FU*3/4
end, MT_DOUBLETEAM)

// Set the Double Team doubles' skin on spawn to prevent errors
addHook("MobjSpawn",function(mo)
	mo.skin = "pikachu"
end,MT_DOUBLETEAM)

addHook("MobjThinker",function(mo)
	mo.alpha = $-(FU/states[mo.state].tics)
end,MT_VOLTTACKLE_SPARK)

// Can't pick up dropped rings after resisting electricity, since otherwise you'd just pick up your lost rings immediately
addHook("TouchSpecial",function(mo,t)
	if t and t.player and t.player.chuTable and t.player.chuTable.zapped
		return true end
end,MT_FLINGRING)

// Reset Pikachu variable tables when changing maps
addHook("MapChange",function()
	for p in players.iterate()
		if p.chuTable
			p.chuTable = nil
		end
	end
end)

// Reset Pikachu variable tables when (re)spawning
addHook("PlayerSpawn",function(p)
	if p.chuTable
		p.chuTable = nil
	end
	
	p.camerascale = skins[p.skin].camerascale
end)